class Configuration:
    __screenshot_path: str = ""
    __save_screenshot_on_fail: bool = True
    __save_full_stacktrace: bool = True

    @property
    def screenshot_path(self) -> str:
        return self.__screenshot_path

    @screenshot_path.setter
    def screenshot_path(self, path: str) -> None:
        self.__screenshot_path = path

    @property
    def save_screenshot_on_fail(self) -> bool:
        return self.__save_screenshot_on_fail

    @save_screenshot_on_fail.setter
    def save_screenshot_on_fail(self, value: bool) -> None:
        self.__save_screenshot_on_fail = value

    @property
    def save_full_stacktrace(self) -> bool:
        return self.__save_full_stacktrace

    @save_full_stacktrace.setter
    def save_full_stacktrace(self, value: bool) -> None:
        self.__save_full_stacktrace = value



import requests


def get_chrome_version(version: str):
    URL = "https://chromedriver.storage.googleapis.com/"
    r = requests.get(url=URL)

    # TODO: it is not pulling the version. I just updated all phones to
    # latest chrome
    chrome_version = ""
    for text in r.text.split(f"{version}."):
        if text.__contains__("/chromedriver_mac64.zip"):
            new_chrome_version = (
                f'{version}.{text.split("/chromedriver_mac64.zip")[0]}'
            )
            if len(new_chrome_version) < 17:
                chrome_version = new_chrome_version
    return chrome_version



import os
from datetime import datetime
from pathlib import Path


class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def log(message, jump_line=False):
    now = datetime.now()
    current_time = now.strftime("%Y-%m-%d %H:%M:%S")
    new_line = ""
    if jump_line:
        new_line = "\n"
    print(f"{new_line}{bcolors.OKBLUE}[{current_time}] {message}{bcolors.ENDC}")
    file = open(__file_log(), "a+")
    file.write(f"{new_line}[{current_time}] {message}\n{new_line}")
    file.close()


def log_pass(message, jump_line=False):
    now = datetime.now()
    current_time = now.strftime("%Y-%m-%d %H:%M:%S")
    new_line = ""
    if jump_line:
        new_line = "\n"
    print(
        f"{new_line}{bcolors.OKGREEN}[{current_time}] {message}{bcolors.ENDC}"
    )
    file = open(__file_log(), "a+")
    file.write(f"{new_line}[{current_time}] {message}\n{new_line}")
    file.close()


def log_error(message, use_date=False):
    now = datetime.now()
    current_time = now.strftime("%Y-%m-%d %H:%M:%S")
    date = ""
    if use_date:
        date = f"[{current_time}] "
    print(f"{bcolors.FAIL}{date}{message}{bcolors.ENDC}")
    file = open(__file_log(), "a+")

    file.write(f"{date}{message}\n")
    file.close()


def log_debug(message):
    now = datetime.now()
    current_time = now.strftime("%Y-%m-%d %H:%M:%S")
    print(f"{bcolors.OKBLUE}[{current_time}] {message}{bcolors.ENDC}")
    file = open(__file_log(), "a+")
    file.write(f"[{current_time}] {message}\n")
    file.close()


def log_warn(message, jump_line=False):
    now = datetime.now()
    current_time = now.strftime("%Y-%m-%d %H:%M:%S")
    new_line = ""
    if jump_line:
        new_line = "\n"
    print(
        f"{new_line}{bcolors.WARNING}[{current_time}] "
        f"{message}{bcolors.ENDC}{new_line}"
    )
    file = open(__file_log(), "a+")
    file.write(f"{new_line}[{current_time}] {message}\n{new_line}")
    file.close()


def log_test_name(message, jump_line=False):
    log_info(message, jump_line)
    new_line = ""
    if jump_line:
        new_line = "\n"
    file = open(__file_tests(), "a+")
    file.write(f"{message}\n{new_line}")
    file.close()


def log_info(message, jump_line=False):
    now = datetime.now()
    current_time = now.strftime("%Y-%m-%d %H:%M:%S")
    new_line = ""
    if jump_line:
        new_line = "\n"
    print(
        f"{new_line}{bcolors.HEADER}[{current_time}] "
        f"{message}{bcolors.ENDC}{new_line}"
    )
    file = open(__file_log(), "a+")
    file.write(f"{new_line}[{current_time}] {message}\n{new_line}")
    file.close()


def __file_log(log_file="stdout.log"):
    root_dir = (
        os.path.dirname(
            os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        )
        + "/"
    )
    Path(root_dir + "appium_logs").mkdir(parents=True, exist_ok=True)
    file_name: str
    if log_file == "stdout.log":
        file_name = f"appium_logs/TEST_UI-{log_file}"
    else:
        file_name = f"appium_logs/{log_file}"
    return root_dir + file_name


def __file_tests(log_file="report_cases.txt"):
    root_dir = (
        os.path.dirname(
            os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        )
        + "/"
    )
    Path(root_dir + "appium_logs").mkdir(parents=True, exist_ok=True)
    return root_dir + log_file



import threading
import time

from testui.elements.testui_element import Elements
from testui.support import logger
from testui.support.configuration import Configuration
from testui.support.helpers import error_with_traceback


class Error(Exception):
    """Base class for exceptions in this module."""

# pylint: disable=super-init-not-called
class CollectionException(Error):
    def __init__(self, message, expression=""):
        self.message = message
        self.expression = expression


class Collections:
    def __init__(self, args):
        self.args = args
        self.__errors = []

    def wait_until_all_visible(self, seconds=10.0, log=True):
        start = time.time()
        threads = []
        for arg in self.args:
            threads.append(
                threading.Thread(
                    target=arg.wait_until_visible, args=(seconds, log)
                )
            )
        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()
        if time.time() < start + seconds:
            logger.log(
                f"{self.args[0].device_name}: Collection of elements has been "
                f"found after {time.time() - start}s"
            )
        else:
            compose_error = ""
            error: str
            for error in self.__errors:
                compose_error = compose_error + f"{error} \n"
            self.__show_error(
                f"{self.args[0].device_name}: Collection of elements has not "
                f"been found after {time.time() - start}s: \n {compose_error}"
            )

    def find_visible(self, seconds=10, return_el_number=False):
        start = time.time()
        arg: Elements
        i = 0
        while time.time() < start + seconds or i < 1:
            for i, arg in enumerate(self.args):
                if arg.is_visible(log=False):
                    logger.log(
                        f"{self.args[i].device_name}: element "
                        f'"{self.args[i].locator_type}: '
                        f'{self.args[i].locator}" found visible'
                    )
                    if return_el_number:
                        return arg, i
                    else:
                        return arg
            i += 1
        self.__show_error(
            f"{self.args[0].device_name}: No element within the collection was "
            f"found visible after {time.time() - start}s:"
        )

    def wait_until_attribute(self, attr_type: list, attr: list, seconds=10):
        start = time.time()
        if len(attr_type) != len(self.args) or len(attr_type) != len(attr):
            raise Exception(
                "The number of attributes checked must be the same as number "
                "of elements in collection"
            )
        threads = []
        i = 0
        element: Elements
        for element in self.args:
            threads.append(
                threading.Thread(
                    target=self.__wait_until_attribute,
                    args=(element, attr_type[i], attr[i], seconds),
                )
            )
            i += 1
        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()
        if time.time() < start + seconds:
            logger.log(
                f"{self.args[0].device_name}: Collection of elements has been "
                "found with the correct attributes "
                f"after {time.time() - start}s"
            )
        else:
            compose_error = ""
            error: str
            for error in self.__errors:
                compose_error = compose_error + f"{error} \n"
            self.__show_error(
                f"{self.args[0].device_name}: Collection of elements has not "
                "been found with the correct attributes "
                f"after {time.time() - start}s: \n {compose_error}"
            )

    def get(self, index: int):
        element: Elements = self.args[index]
        return element

    def __wait_until_attribute(
        self, element: Elements, attr_type, attr, seconds
    ):
        try:
            element.wait_until_attribute(attr_type, attr, seconds)
        except Exception as err:
            self.__errors.append(err)

    def __show_error(self, exception) -> None:
        driver = self.args[0].testui_driver
        config: Configuration = driver.configuration

        if config.save_screenshot_on_fail:
            driver.save_screenshot()

        full_exception = exception
        if config.save_full_stacktrace:
            full_exception = error_with_traceback(exception)

        if driver.soft_assert:
            logger.log_error(full_exception)
            driver.set_error(full_exception)
        else:
            logger.log_error(full_exception)
            driver.set_error(full_exception)
            raise CollectionException(exception)


def ee(*args) -> Collections:
    """
    locator types:
        id,
        css,
        className,
        name,
        xpath,
        accessibility,
        uiautomator,
        classChain,
        predicate
    """
    return Collections(args)



import base64
import os

from datetime import datetime
from os import path
from pathlib import Path

from appium.webdriver.common.touch_action import TouchAction
from appium.webdriver.webdriver import WebDriver
from selenium.webdriver import ActionChains
from selenium.common.exceptions import WebDriverException

from testui.elements.testui_element import e
from testui.support import logger
from testui.support.helpers import error_with_traceback
from testui.support.testui_images import get_point_match, ImageRecognition
from testui.support.configuration import Configuration


class TestUIDriver:
    __test__ = False

    def __init__(self, driver):
        self.__soft_assert = driver.soft_assert
        self.__appium_driver = driver.get_driver()
        self.__process = driver.process
        self.errors = []
        self.appium_port = driver.appium_port
        self.browser: bool = driver.browser
        self.logger_name = driver.logger_name
        self.soft_assert: bool = driver.soft_assert
        self.device_udid = driver.udid
        self.device_name = driver.device_name
        self.file_name = driver.file_name
        self.__configuration: Configuration = driver.configuration

    def switch_to_context(self, context=0, last=False):
        if last:
            context = len(self.__appium_driver.contexts) - 1
        try:
            if len(self.__appium_driver.contexts) == 1:
                logger.log(
                    f"{self.device_name}: There is only one context: "
                    f"{self.__appium_driver.contexts[context]}"
                )
            elif context >= len(self.__appium_driver.contexts):
                logger.log_warn(
                    f"{self.device_name}: Cannot switch to context {context}: "
                    f"there are just {len(self.__appium_driver.contexts)} "
                    "contexts"
                )
            self.__appium_driver.execute(
                "switchToContext",
                {"name": self.__appium_driver.contexts[context]},
            )
            logger.log(
                f"{self.device_name}: Switched to context: "
                f"{self.__appium_driver.contexts[context]}"
            )
        except Exception as err:
            if self.__soft_assert:
                logger.log_error(err)
            else:
                raise err
        return self

    @property
    def context(self):
        return self.__appium_driver.contexts

    def e(self, locator_type, locator):
        return e(self, locator_type, locator)

    def execute(self, driver_command, params=None):
        self.get_driver().execute(driver_command, params)

    def remove_log_file(self, when_no_errors=True):
        """
        removes appium log file. If when_no_errors is False, it will always
        remove errors, if True then just when there are errors.
        :param when_no_errors:
        :return:
        """
        if self.file_name is not None:
            try:
                if when_no_errors and len(self.errors) == 0:
                    os.remove(self.file_name)
                elif not when_no_errors:
                    os.remove(self.file_name)
            except FileNotFoundError:
                logger.log_debug("Log file already removed")

    def touch_actions(self):
        """
        This method is meant for Appium Drivers Only
        :return:
        """
        return TouchAction(self.get_driver())

    def actions(self):
        """
        This method is meant for Selenium or Browser views (Mobile and Desktop)
        :return:
        """
        return ActionChains(self.get_driver())

    def open_notifications(self):
        self.get_driver().open_notifications()

    def back(self):
        self.get_driver().back()
        return self

    def quit(self, stop_server=True):
        self.get_driver().quit()
        if self.__process is not None and stop_server:
            self.__process.kill()

    def navigate_to(self, url):
        self.get_driver().get(url)
        logger.log(f"{self.device_name}: Navigating to: {url}")

    def execute_script(self, driver_command, args: None):
        self.get_driver().execute_script(driver_command, args)

    @property
    def switch_to(self):
        return self.get_driver().switch_to

    def set_network_connection(self, number):
        self.__appium_driver.set_network_connection(number)

    @property
    def network_connection(self):
        return self.__appium_driver.network_connection

    def find_image_match(
        self,
        comparison,
        threshold=0.90,
        assertion=False,
        not_found=False,
        image_match="",
    ):
        now = datetime.now()
        current_time = now.strftime("%Y-%m-%d%H%M%S")
        image_name = f"{self.device_udid}{current_time}.png"
        image_path = self.save_screenshot(image_name)
        found, p = ImageRecognition(
            image_path, comparison, threshold, self.device_name
        ).compare(image_match)
        if assertion and not found and not not_found:
            exception = self.new_error_message(
                "The images compared did not match"
                f"Threshold={threshold}, matched = {p}"
            )
            logger.log_error(error_with_traceback(exception))
            raise Exception(exception)
        os.remove(image_path)

        return found

    def click_by_image(self, image: str, threshold=0.9, webview=False):
        now = datetime.now()
        current_time = now.strftime("%Y-%m-%d%H%M%S")
        image_name = f"{self.device_udid}{current_time}.png"
        self.save_screenshot(image_name)
        x, y = get_point_match(
            f"testui-{image_name}", f"{image}", threshold, self.device_name
        )
        ta = TouchAction(self.__appium_driver)
        if webview:
            y = y + 120
        ta.tap(x=x, y=y).perform()
        logger.log(f"{self.device_name}: element with image {image} clicked")
        self.__delete_screenshot(image_name)

    def get_dimensions(self):
        now = datetime.now()
        current_time = now.strftime("%Y-%m-%d%H%M%S")
        image_name = f"{self.device_udid}{current_time}.png"
        self.save_screenshot(image_name)
        dimensions = ImageRecognition(
            f"testui-{image_name}"
        ).image_original_size()
        self.__delete_screenshot(image_name)
        return dimensions

    def click(self, x, y):
        ta = TouchAction(self.__appium_driver)
        ta.tap(x=x, y=y).perform()
        logger.log(f'Clicked over "x={x}: y={y}"')

    def save_screenshot(self, image_name=""):
        config = self.__configuration

        root_dir = config.screenshot_path
        if not config.screenshot_path:
            root_dir = path.dirname(
                path.dirname(path.dirname(path.abspath(__file__)))
            )
            root_dir = path.join(root_dir, "report_screenshots")

        Path(root_dir).mkdir(parents=True, exist_ok=True)

        current_time = datetime.now().strftime("%Y-%m-%d%H%M%S")

        if not image_name:
            image_name = f"ERROR-{self.device_name}-{current_time}.png"

        final_path = path.join(root_dir, image_name)

        self.get_driver().save_screenshot(final_path)

        logger.log_debug(
            self.new_error_message(f'Screenshot saved in "{final_path}"')
        )

        return final_path

    @classmethod
    def __delete_screenshot(cls, image_name):
        root_dir = os.path.dirname(
            os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        )
        os.remove(root_dir + f"/testui-{image_name}")

    def get_driver(self) -> WebDriver:
        driver = self.__appium_driver

        return driver

    @property
    def configuration(self) -> Configuration:
        return self.__configuration

    def set_error(self, error):
        self.errors.append(error)

    def raise_errors(self, remove_log_file=False):
        if len(self.errors) != 0:
            composed_error = "\n"
            i = 1
            for error in self.errors:
                composed_error += f"Error {i.__str__()}: {error}\n"
                i += 1
            self.errors = []
            raise Exception(composed_error)
        if remove_log_file:
            self.remove_log_file()

    def get_clipboard_text(self) -> str:
        return self.__appium_driver.get_clipboard_text()

    def set_power_capacity(self, capacity: int):
        try:
            self.get_driver().set_power_capacity(capacity)
        except WebDriverException as wd_exception:
            exception = self.new_error_message(
                "powerCapacity method is only available for emulators"
            )
            logger.log_error(error_with_traceback(exception))
            raise Exception(exception) from wd_exception

    def background_app(self, seconds):
        self.get_driver().background_app(seconds)

    def remove_app(self, app_id):
        self.get_driver().remove_app(app_id)

    def install_app(self, app_id):
        self.get_driver().install_app(app_id)

    def start_recording_screen(self):
        self.get_driver().start_recording_screen()

    def stop_recording_screen(self, file_name="testui-video.mp4"):
        file = self.get_driver().stop_recording_screen()
        decoded_string = base64.b64decode(file)
        root_dir = os.path.dirname(
            os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        )
        with open(root_dir + f"/{file_name}", "wb") as wfile:
            wfile.write(decoded_string)

    def stop_recording_and_compare(
        self,
        comparison,
        threshold=0.9,
        fps_reduction=1,
        not_found=False,
        keep_image_as="",
        assertion=True,
    ):
        now = datetime.now()
        current_time = now.strftime("%Y-%m-%d%H%M%S")
        video_name = f"{self.device_udid}{current_time}.mp4"
        self.stop_recording_screen(video_name)
        root_dir = os.path.dirname(
            os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        )
        found = ImageRecognition(
            video_name, comparison, threshold, device_name=self.device_name
        ).compare_video(keep_image_as, frame_rate_reduction=fps_reduction)

        os.remove(root_dir + f"/{video_name}")
        if not found and not not_found:
            if assertion:
                raise Exception(
                    self.new_error_message(
                        "The image and video compared did not match. "
                        f"Threshold={threshold}"
                    )
                )

            return False
        if found and not_found:
            if assertion:
                raise Exception(
                    self.new_error_message(
                        "The image and video compared matched. "
                        f"Threshold={threshold}"
                    )
                )

            return False
        return True

    def new_error_message(self, message) -> str:
        return f"{self.device_name}: {message}"

    def hide_keyboard(self):
        self.get_driver().hide_keyboard()



import atexit
import os
import subprocess
import threading
from pathlib import Path
from time import sleep

import geckodriver_autoinstaller

from ppadb.client import Client as AdbClient
from appium.webdriver import Remote
from appium.webdriver.webdriver import WebDriver
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.firefox.options import Options as FirefoxOptions
from selenium import webdriver
from webdriver_manager import chrome

from testui.support import logger
from testui.support.api_support import get_chrome_version
from testui.support.testui_driver import TestUIDriver
from testui.support.configuration import Configuration


class NewDriver:
    __configuration = Configuration()

    def __init__(self):
        self.browser = False
        self.__driver: WebDriver = None
        self.__app_path = None
        self.udid = None
        self.__appium_url = None
        self.__remote_url = None
        self.__browser_name = "chrome"
        self.device_name = "Device"
        self.appium_port = 4723
        self.__version = None
        self.__platform_name = "Android"
        self.__app_package = None
        self.__app_activity = None
        self.__automation_name = None
        self.logger_name = None
        self.__full_reset = False
        self.__debug = False
        self.soft_assert = False
        # TODO: Investigate if functionality can be implemented or should be
        # removed.
        # pylint: disable=unused-private-member
        self.__auto_accept_alerts = True
        self.process = None
        self.file_name = None
        self.__appium_log_file = "appium-stdout.log"
        self.__chromedriverArgs = ["relaxed security"]
        self.__desired_capabilities = {}
        # TODO: Investigate if should be used in functionality or should be
        # removed.
        # pylint: disable=unused-private-member
        self.__chrome_options = {}

    def set_logger(self, logger_name: str or None = "pytest"):
        """Possible loggers str: behave, pytest, None"""
        self.logger_name = logger_name
        return self

    def set_appium_log_file(self, file="appium-stdout.log"):
        self.__appium_log_file = file
        return self

    def set_browser(self, browser: str) -> "NewDriver":
        self.__browser_name = browser
        return self

    def set_remote_url(self, url):
        self.__remote_url = url
        return self

    def set_soft_assert(self, soft_assert: bool):
        self.soft_assert = soft_assert
        return self

    def set_appium_port(self, port: int):
        self.appium_port = port
        return self

    def set_full_reset(self, full_reset: bool):
        self.__full_reset = full_reset
        return self

    def set_appium_url(self, appium_url: str):
        self.__appium_url = appium_url
        return self

    def set_extra_caps(self, caps=None):
        if caps is None:
            caps = {}
        for cap in caps:
            self.__desired_capabilities[cap] = caps[cap]
        return self

    def set_app_path(self, path: str):
        self.__app_path = path
        if os.path.isabs(self.__app_path):
            return self
        else:
            root_dir = os.path.dirname(
                os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
            )
            self.__app_path = os.path.join(root_dir, path)
            logger.log(self.__app_path)
            return self

    def set_udid(self, udid: str):
        self.udid = udid
        return self

    def set_udid_if_exists(self, udid: str, number=None):
        self.udid = check_device_exist(udid)
        if self.udid is None:
            self.udid = get_device_udid(number)
        return self

    def set_connected_device(self, number: int):
        self.udid = get_device_udid(number)
        return self

    def set_device_name(self, device_name: str):
        self.device_name = device_name
        return self

    def set_version(self, version: str):
        self.__version = version
        return self

    def set_grant_permissions(self, permissions: bool):
        # pylint: disable=unused-private-member
        self.__auto_accept_alerts = permissions
        return self

    def set_app_package_activity(self, app_package: str, app_activity: str):
        self.__app_package = app_package
        self.__app_activity = app_activity
        return self

    def get_driver(self) -> WebDriver:
        driver = self.__driver
        return driver

    @property
    def configuration(self) -> Configuration:
        return self.__configuration

    def get_testui_driver(self) -> TestUIDriver:
        return TestUIDriver(self)

    def set_chrome_driver(self, version="") -> "NewDriver":
        mobile_version = version
        if version == "":
            if self.udid is None:
                self.udid = get_device_udid(0)
            mobile_version = check_chrome_version(self.udid)
        chrome_driver = chrome.ChromeDriverManager(
            version=mobile_version
        ).install()
        logger.log(f"Driver installed in {chrome_driver}", True)
        self.__desired_capabilities["chromedriverExecutable"] = chrome_driver
        return self

    def set_screenshot_path(self, screenshot_path: str):
        self.__configuration.screenshot_path = screenshot_path
        return self

    def set_save_screenshot_on_fail(self, save_screenshot_on_fail: bool):
        self.__configuration.save_full_stacktrace = save_screenshot_on_fail
        return self

    def set_save_full_stacktrace(self, save_full_stacktrace: bool):
        self.__configuration.save_full_stacktrace = save_full_stacktrace
        return self

    # Available platforms: Android, iOS
    def set_platform(self, platform):
        self.__platform_name = platform
        return self

    def __set_common_caps(self):
        self.__desired_capabilities["adbExecTimeout"] = 30000
        self.__desired_capabilities["platformName"] = self.__platform_name
        self.__desired_capabilities["automationName"] = self.__automation_name
        self.__desired_capabilities["deviceName"] = self.device_name
        if self.__full_reset:
            self.__desired_capabilities["enforceAppInstall"] = True
        else:
            self.__desired_capabilities["noReset"] = True
        if self.__version is not None:
            self.__desired_capabilities["platformVersion"] = self.__version
        if self.udid is not None:
            self.__desired_capabilities["udid"] = self.udid

    def __set_android_caps(self):
        if self.__automation_name is None:
            self.__automation_name = "UiAutomator2"
        self.__desired_capabilities["chromeOptions"] = {"w3c": False}
        # TODO: It is not being passed to executable. Tried this
        # https://github.com/appium/appium/blob/master/docs/en/writing-running-appium/caps.md
        self.__desired_capabilities[
            "chromedriverArgs"
        ] = self.__chromedriverArgs

        self.__desired_capabilities["chromeDriverPort"] = (
            self.appium_port - 4723 + 8100
        )
        self.__desired_capabilities["systemPort"] = (
            self.appium_port - 4723 + 8200
        )
        if self.__app_path is None and self.__app_package is None:
            self.__desired_capabilities["browserName"] = "chrome"
            self.browser = True
        if self.__app_package is not None:
            self.__desired_capabilities["appPackage"] = self.__app_package
            self.__desired_capabilities["appActivity"] = self.__app_activity
        if self.__app_path is not None:
            self.__desired_capabilities["app"] = self.__app_path
            self.__desired_capabilities["androidInstallPath"] = self.__app_path

    def __set_ios_caps(self):
        if self.__automation_name is None:
            self.__automation_name = "XCUITest"
        if self.__app_path is None and self.__app_package is None:
            self.__desired_capabilities["browserName"] = "safari"
            self.browser = True
        if self.__app_path is not None:
            self.__desired_capabilities["app"] = self.__app_path
        if self.__version is None:
            self.__desired_capabilities["platformVersion"] = "13.2"

    def __set_selenium_caps(self):
        self.__desired_capabilities["browserName"] = self.__browser_name

    def set_appium_driver(self) -> TestUIDriver:
        if self.__platform_name.lower() == "android":
            self.__set_android_caps()
        else:
            self.__set_ios_caps()
        self.__set_common_caps()
        self.__driver, self.process, self.file_name = start_driver(
            self.__desired_capabilities,
            self.__appium_url,
            self.__debug,
            self.appium_port,
            self.udid,
            self.__appium_log_file,
        )
        return self.get_testui_driver()

    def set_selenium_driver(
        self,
        chrome_options: ChromeOptions or None = None,
        firefox_options: FirefoxOptions or None = None,
    ) -> TestUIDriver:
        self.__set_selenium_caps()
        self.__driver = start_selenium_driver(
            self.__desired_capabilities,
            self.__remote_url,
            self.__debug,
            self.__browser_name,
            chrome_options,
            firefox_options,
        )

        return self.get_testui_driver()

    def set_driver(self, driver) -> TestUIDriver:
        self.__set_selenium_caps()
        self.__driver = driver
        return self.get_testui_driver()


def start_driver(desired_caps, url, debug, port, udid, log_file):
    lock = threading.Lock()
    lock.acquire()

    logger.log("setting capabilities: " + desired_caps.__str__())
    logger.log("starting appium driver...")

    process = None
    if desired_caps["platformName"].lower().__contains__("android"):
        url, desired_caps, process, file = __local_run(
            url, desired_caps, port, udid, log_file
        )
    else:
        url, desired_caps, file = __local_run_ios(
            url, desired_caps, port, udid, log_file
        )
    err = None
    for _ in range(2):
        try:
            driver = Remote(url, desired_caps)
            atexit.register(__quit_driver, driver, debug)
            logger.log(f"appium running on {url}. \n")
            lock.release()
            return driver, process, file
        except Exception as error:
            err = error
    lock.release()
    raise err


def start_selenium_driver(
    desired_caps,
    url=None,
    debug=None,
    browser: str or None = None,
    chrome_options: ChromeOptions or None = None,
    firefox_options: FirefoxOptions or None = None,
) -> WebDriver:
    """Starts a new local session of the specified browser."""

    options = chrome_options
    if firefox_options is not None:
        options = firefox_options

    if options is not None:
        logger.log(f"setting options: {options.to_capabilities().__str__()}")

    logger.log(f"setting capabilities: {desired_caps.__str__()}")
    logger.log(f"starting selenium {browser.lower()} driver...")

    err = None
    for _ in range(2):
        try:
            if url is not None:
                logger.log(f"selenium running on {url}. \n")

                for key, value in desired_caps.items():
                    options.set_capability(key, value)

                driver = webdriver.Remote(command_executor=url, options=options)
            else:
                if browser.lower() == "chrome":
                    for key, value in desired_caps.items():
                        options.set_capability(key, value)
                    driver = webdriver.Chrome(options=options)
                elif browser.lower() == "firefox":
                    try:
                        geckodriver_autoinstaller.install()
                    except Exception as error:
                        logger.log_warn(
                            "Could not retrieve geckodriver: " + error.__str__()
                        )
                    if "marionette" not in desired_caps:
                        desired_caps["marionette"] = True
                    driver = webdriver.Firefox(
                        desired_capabilities=desired_caps, options=options
                    )
                elif browser.lower() == "safari":
                    driver = webdriver.Safari(desired_capabilities=desired_caps)
                elif browser.lower() == "edge":
                    driver = webdriver.Edge(capabilities=desired_caps)
                elif browser.lower() == "ie":
                    driver = webdriver.Ie(capabilities=desired_caps)
                elif browser.lower() == "opera":
                    driver = webdriver.Opera(desired_capabilities=desired_caps)
                else:
                    raise Exception(
                        f"Invalid browser '{browser}'. Please choose one "
                        f"from: chrome, firefox, safari, edge, ie, opera"
                    )
            atexit.register(__quit_driver, driver, debug)
            return driver
        except Exception as error:
            err = error

    raise err


def __local_run(url, desired_caps, use_port, udid, log_file):
    if url is None:
        port = use_port
        bport = use_port + 1
        device = 0
        if os.getenv("PYTEST_XDIST_WORKER") is not None:
            device += os.getenv("PYTEST_XDIST_WORKER").split("w")[1]
            port += int(os.getenv("PYTEST_XDIST_WORKER").split("w")[1]) * 2
            desired_caps["chromeDriverPort"] = 8200 + int(
                os.getenv("PYTEST_XDIST_WORKER").split("w")[1]
            )
            desired_caps["systemPort"] = 8300 + int(
                os.getenv("PYTEST_XDIST_WORKER").split("w")[1]
            )
            bport += int(os.getenv("PYTEST_XDIST_WORKER").split("w")[1]) * 2
        logger.log(f"running: appium -p {port.__str__()} -bp {bport.__str__()}")
        if udid is None:
            desired_caps = __set_android_device(desired_caps, device)
        logger.log(f'setting device for automation: {desired_caps["udid"]}')
        root_dir = (
            os.path.dirname(
                os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
            )
            + "/"
        )
        Path(root_dir + "appium_logs").mkdir(parents=True, exist_ok=True)
        file_path: str
        if log_file == "appium-stdout.log":
            file = f"appium_logs/testui-{udid}-" + log_file
        else:
            file = f"appium_logs/{log_file}"
        with open(root_dir + file, "wb") as out:
            process = subprocess.Popen(
                ["appium", "-p", port.__str__(), "-bp", bport.__str__()],
                stdout=out,
                stderr=subprocess.STDOUT,
            )
            atexit.register(process.kill)
        file_path = root_dir + file
        while True:
            sleep(0.5)
            out = open(file_path)
            text = out.read()
            if text.__contains__("already be in use") or text.__contains__(
                "listener started"
            ):
                out.close()
                break
            out.close()
        return (
            f"http://localhost:{port.__str__()}/wd/hub",
            desired_caps,
            process,
            file_path,
        )
    return url, desired_caps, None


def __local_run_ios(url, desired_caps, use_port, udid, log_file):
    process = None
    if url is None:
        port = use_port + 100
        device = 0
        if os.getenv("PYTEST_XDIST_WORKER") is not None:
            device += os.getenv("PYTEST_XDIST_WORKER").split("w")[1]
            port += int(os.getenv("PYTEST_XDIST_WORKER").split("w")[1]) * 2
            desired_caps["chromeDriverPort"] = 8200 + int(
                os.getenv("PYTEST_XDIST_WORKER").split("w")[1]
            )
            desired_caps["systemPort"] = 8300 + int(
                os.getenv("PYTEST_XDIST_WORKER").split("w")[1]
            )
        logger.log(f"running: appium -p {port.__str__()}")
        root_dir = (
            os.path.dirname(
                os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
            )
            + "/"
        )
        Path(root_dir + "appium_logs").mkdir(parents=True, exist_ok=True)
        file_path: str
        if log_file == "appium-stdout.log":
            file = f"appium_logs/testui-{udid}-" + log_file
        else:
            file = f"appium_logs/{log_file}"
        with open(root_dir + file, "wb") as out:
            process = subprocess.Popen(
                ["appium", "-p", port.__str__()],
                stdout=out,
                stderr=subprocess.STDOUT,
            )
            atexit.register(process.kill)
        file_path = root_dir + file
        if udid is None:
            desired_caps = __set_ios_device(desired_caps, device)
        while True:
            sleep(0.5)
            out = open(file_path)
            text = out.read()
            if text.__contains__("already be in use") or text.__contains__(
                "listener started"
            ):
                out.close()
                break
            out.close()
        return (
            f"http://localhost:{port.__str__()}/wd/hub",
            desired_caps,
            file_path,
        )
    return url, desired_caps, process


def __set_android_device(desired_caps, number: int):
    desired_caps["udid"] = get_device_udid(number)
    return desired_caps


def __set_ios_device(desired_caps, number: int):
    # TODO implement function
    _ = number

    return desired_caps


def get_device_udid(number: int):
    client = AdbClient(host="127.0.0.1", port=5037)
    devices = client.devices()
    if len(devices) == 0:
        raise Exception("There are 0 devices connected to the computer!")
    if len(devices) > number:
        return devices[number].get_serial_no()
    else:
        new_number = number - (number // len(devices)) * len(devices)
        logger.log_warn(
            f"You choose device number {number + 1} but there are only "
            f"{len(devices)} connected. "
            f"Will use device number {new_number + 1} instead",
            jump_line=True,
        )
        return devices[new_number].get_serial_no()


def check_device_exist(udid):
    client = AdbClient(host="127.0.0.1", port=5037)
    devices = client.devices()
    for device in devices:
        if device.get_serial_no() == udid:
            return udid
    return None


def check_chrome_version(udid):
    output = subprocess.Popen(
        [
            "adb",
            "-s",
            udid,
            "shell",
            "dumpsys",
            "package",
            "com.android.chrome",
            "|",
            "grep",
            "versionName",
        ],
        stdout=subprocess.PIPE,
    )
    response = output.communicate()
    if response.__str__().__contains__("versionName="):
        return get_chrome_version(
            response.__str__().split("versionName=")[1].split(".")[0]
        )

    return None


def __quit_driver(driver, debug):
    try:
        driver.quit()
    except Exception as err:
        if debug:
            logger.log_debug(f"appium was probably closed {err}. \n")



import os
import threading
import time

import cv2
import numpy as np
import imutils

from testui.support import logger

found_image = False
matched = 0.0


def compare_video_image(
    video,
    comparison,
    threshold,
    image_match,
    frame_rate_reduction=1,
    max_scale=2.0,
):
    root_dir = os.path.dirname(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    )
    cap = cv2.VideoCapture(root_dir + "/" + video)
    template = cv2.imread(root_dir + "/" + comparison)
    i = 0
    percentage = 0.0
    while cap.isOpened():
        # Capture frame-by-frame
        ret, frame = cap.read()
        if ret and i % frame_rate_reduction == 0:
            found, percentage = __compare(
                frame, template, threshold, image_match, root_dir, max_scale
            )
            if found:
                cap.release()
                logger.log(f"Match found in the {i}th frame of the video")
                return True, percentage
        elif not ret:
            break
        # sleep(frame_rate_reduction)
        i += 1
    cap.release()
    return False, percentage


def __compare(
    image,
    template,
    threshold: float,
    image_match: str,
    root_dir: str,
    max_scale: float,
    min_scale=0.1,
):
    (tH, tW) = template.shape[:2]
    # loop over the scales of the image
    found = None
    global found_image
    global matched
    maxVal = 0.0
    for scale in np.linspace(min_scale, max_scale, 5)[::-1]:
        # resize the image according to the scale, and keep track of the ratio
        # of the resizing.
        resized = imutils.resize(image, width=int(image.shape[1] * scale))
        r = image.shape[1] / float(resized.shape[1])
        # if the resized image is smaller than the template, then break from
        # the loop
        if resized.shape[0] < tH or resized.shape[1] < tW:
            break
        result = cv2.matchTemplate(resized, template, cv2.TM_CCOEFF_NORMED)
        (_, maxVal, _, maxLoc) = cv2.minMaxLoc(result)
        # if we have found a new maximum correlation value, then update the
        # bookkeeping variable
        if found is None or maxVal > found[0]:
            lock = threading.Lock()
            lock.acquire()
            if found_image:
                lock.release()
                return True, matched
            lock.release()
            found = (maxVal, maxLoc, r)
            if maxVal > threshold:
                if image_match != "" and found is not None:
                    # unpack the bookkeeping variable and compute the (x, y)
                    # coordinates of the bounding box based on the resized ratio
                    (_, maxLoc, r) = found
                    (startX, startY) = (int(maxLoc[0] * r), int(maxLoc[1] * r))
                    (endX, endY) = (
                        int((maxLoc[0] + tW) * r),
                        int((maxLoc[1] + tH) * r),
                    )
                    # draw a bounding box around the detected result and display
                    # the image
                    cv2.rectangle(
                        image, (startX, startY), (endX, endY), (0, 0, 255), 2
                    )
                    cv2.imwrite(os.path.join(root_dir, image_match), image)
                lock.acquire()
                found_image = True
                lock.release()
                matched = maxVal
                return True, maxVal
    matched = maxVal
    return False, maxVal


def compare_images(
    original: str,
    comparison: str,
    threshold=0.9,
    image_match="",
    max_scale=2.0,
    min_scale=0.3,
):
    # Read the images from the file
    global found_image
    global matched

    start = time.time()
    matched = 0.0
    found_image = False
    root_dir = os.path.dirname(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    )
    if not os.path.exists(comparison):
        comparison = os.path.join(root_dir, comparison)
        if not os.path.exists(comparison):
            raise Exception(f"There is no image in {comparison}")
    if not os.path.exists(original):
        original = os.path.join(root_dir, original)
        if not os.path.exists(original):
            raise Exception(f"There is no image in {original}")
    template = cv2.imread(comparison)
    image = cv2.imread(original)
    # loop over the scales of the image
    threads = []
    parts = max_scale / 5.0
    min_scale = min(min_scale, parts)

    threads.append(
        threading.Thread(
            target=__compare,
            args=(
                image,
                template,
                threshold,
                image_match,
                root_dir,
                parts,
                min_scale,
            ),
        )
    )
    threads.append(
        threading.Thread(
            target=__compare,
            args=(
                image,
                template,
                threshold,
                image_match,
                root_dir,
                parts * 2.0,
                parts,
            ),
        )
    )
    threads.append(
        threading.Thread(
            target=__compare,
            args=(
                image,
                template,
                threshold,
                image_match,
                root_dir,
                parts * 3.0,
                parts * 2.0,
            ),
        )
    )
    threads.append(
        threading.Thread(
            target=__compare,
            args=(
                image,
                template,
                threshold,
                image_match,
                root_dir,
                parts * 4.0,
                parts * 3.0,
            ),
        )
    )
    threads.append(
        threading.Thread(
            target=__compare,
            args=(
                image,
                template,
                threshold,
                image_match,
                root_dir,
                parts * 5.0,
                parts * 4.0,
            ),
        )
    )
    for thread in threads:
        thread.start()
    while not found_image:
        alive = False
        for thread in threads:
            if thread.is_alive():
                alive = True
        if not alive:
            break

    logger.log(f"Image recognition took {time.time() - start}s")
    return found_image, matched


def get_point_match(
    original: str, comparison: str, threshold=0.9, device_name="Device"
):
    _ = device_name

    # Read the images from the file
    root_dir = os.path.dirname(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    )
    template = cv2.imread(root_dir + "/" + comparison)
    (tH, tW) = template.shape[:2]
    image = cv2.imread(root_dir + "/" + original)
    found = None
    # loop over the scales of the image
    for scale in np.linspace(0.2, 1.0, 10)[::-1]:
        # resize the image according to the scale, and keep track of the ratio
        # of the resizing
        resized = imutils.resize(image, width=int(image.shape[1] * scale))
        r = image.shape[1] / float(resized.shape[1])
        # if the resized image is smaller than the template, then break from
        # the loop
        if resized.shape[0] < tH or resized.shape[1] < tW:
            break
        result = cv2.matchTemplate(resized, template, cv2.TM_CCOEFF_NORMED)
        (_, maxVal, _, maxLoc) = cv2.minMaxLoc(result)
        # if we have found a new maximum correlation value, then update the
        # bookkeeping variable
        if found is None or maxVal > found[0]:
            found = (maxVal, maxLoc, r)
            if maxVal > threshold:
                break

    # unpack the bookkeeping variable and compute the (x, y) coordinates
    # of the bounding box based on the resized ratio
    (_, maxLoc, r) = found
    (startX, startY) = (int(maxLoc[0] * r), int(maxLoc[1] * r))
    (endX, endY) = (int((maxLoc[0] + tW) * r), int((maxLoc[1] + tH) * r))

    return startX + (endX - startX) // 2, startY + (endY - startY) // 2


def draw_match(
    original: str, comparison: str, threshold=0.9, device_name="Device"
):
    method = cv2.TM_CCOEFF_NORMED

    # Read the images from the file
    root_dir = os.path.dirname(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    )
    large_image = cv2.imread(root_dir + "/" + comparison)
    small_image = cv2.imread(root_dir + "/" + original)

    logger.log_debug(
        f'{device_name}: Comparing "{original}" with "{comparison}"'
    )
    res = cv2.matchTemplate(small_image, large_image, method)

    loc = np.where(res >= threshold)
    _, w, h = large_image.shape[::-1]
    # For each match...
    suh = None
    for pt in zip(*loc[::-1]):
        suh = cv2.rectangle(
            small_image, pt, (pt[0] + w, pt[1] + h), (0, 66, 255), 1
        )
    cv2.imwrite("something.png", suh)


def size(image_path):
    root_dir = os.path.dirname(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    )
    img = cv2.imread(root_dir + "/" + image_path)
    height, width, _ = img.shape
    return width, height


class ImageRecognition:
    def __init__(
        self, original: str, comparison="", threshold=0.9, device_name="Device"
    ):
        self.__original = original
        self.__comparison = comparison
        self.__threshold = threshold
        self.__device_name = device_name

    def compare(self, image_match="", max_scale=2.0, min_scale=0.3):
        _, p1 = compare_images(
            self.__original,
            self.__comparison,
            self.__threshold,
            image_match,
            max_scale,
            min_scale,
        )
        if self.__threshold > p1:
            logger.log_debug(
                f"{self.__device_name}: Image match not found between: "
                f"{self.__original} and {self.__comparison}. "
                f"Threshold={self.__threshold}, matched = {p1}"
            )
            return False, p1

        logger.log_debug(
            f"{self.__device_name}: Image match found between: "
            f"{self.__original} and {self.__comparison}. "
            f"Threshold={self.__threshold}, matched = {p1}"
        )
        return True, p1

    def compare_video(
        self, image_match="", frame_rate_reduction=1, max_scale=2.0
    ):
        found, p = compare_video_image(
            self.__original,
            self.__comparison,
            self.__threshold,
            image_match,
            frame_rate_reduction,
            max_scale,
        )
        if found:
            logger.log_debug(
                f"{self.__device_name}: Image match found between "
                f"video: {self.__original} and image {self.__comparison}. "
                f"Threshold={self.__threshold}, matched = {p}"
            )
            return True

        logger.log_debug(
            f"{self.__device_name}: Image match not found between "
            f"video: {self.__original} and image {self.__comparison}. "
            f"Threshold={self.__threshold}, matched = {p}"
        )
        return False

    def get_middle_point(self):
        get_point_match(
            self.__original,
            self.__comparison,
            self.__threshold,
            self.__device_name,
        )
        return self

    def draw_image_match(self):
        draw_match(
            self.__original,
            self.__comparison,
            self.__threshold,
            self.__device_name,
        )
        return self

    def image_original_size(self):
        size_image = size(self.__original)
        logger.log(f"The size of the image is {size_image}")
        return Dimensions(size_image[0], size_image[1])

    def image_comparison_size(self):
        size_image = size(self.__comparison)
        logger.log(f"The size of the image is {size_image}")
        return Dimensions(size_image[0], size_image[1])

    def crop_original_image(
        self, center_x, center_y, width, height, image_name="cropped_image.png"
    ):
        # Read the images from the file
        root_dir = os.path.dirname(
            os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        )
        img = cv2.imread(root_dir + "/" + self.__original)
        y = center_y - height // 2
        if y < 0:
            y *= -1
        x = center_x - width // 2
        if x < 0:
            x *= -1
        img_2 = img[y : y + height, x : x + width]
        cv2.imwrite(root_dir + "/" + image_name, img_2)
        return self


class Dimensions:
    def __init__(self, x, y):
        self.x = x
        self.y = y



import os
import traceback

from testui.support import logger


def error_with_traceback(exception):
    root_dir = os.path.dirname(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    )
    line: str
    for line in traceback.extract_stack().format():
        if line.__contains__(root_dir) and not line.__contains__(
            "traceback.extract_stack()"
        ):
            exception += logger.bcolors.FAIL + line + logger.bcolors.ENDC
    return exception



from __future__ import annotations

import time
import os

from os import path
from typing import List

from appium.webdriver.common.touch_action import TouchAction
from appium.webdriver.webelement import WebElement
from selenium.webdriver import ActionChains
from selenium.webdriver.common.by import By

from testui.support import logger
from testui.support.helpers import error_with_traceback
from testui.support.testui_images import Dimensions, ImageRecognition

def testui_error(driver, exception: str) -> None:
    config = driver.configuration

    if config.save_screenshot_on_fail:
        try:
            driver.save_screenshot()
        except Exception as error:
            exception += (
                f"{logger.bcolors.FAIL} \n"
                f"Could not take screenshot:{logger.bcolors.ENDC}\n {error}"
            )

    full_exception = exception
    if config.save_full_stacktrace:
        full_exception = error_with_traceback(exception)

    if driver.soft_assert:
        logger.log_error(full_exception)
        driver.set_error(full_exception)
    else:
        logger.log_error(full_exception)
        driver.set_error(full_exception)
        raise ElementException(
            "There were errors during the UI testing, check the logs"
        )


class Error(Exception):
    """Base class for exceptions in this module."""


# pylint: disable=super-init-not-called
class ElementException(Error):
    def __init__(self, message, expression=""):
        self.message = message
        self.expression = expression


class Elements:
    def __init__(self, driver, locator_type: str, locator: str):
        self.logger = driver.logger_name
        # TODO: Investigate if should be used in functionality or should be
        # removed.
        # pylint: disable=unused-private-member
        self.__soft_assert = driver.soft_assert
        self.testui_driver = driver
        self.device_name = driver.device_name
        self.driver = driver.get_driver()
        self.locator = locator
        self.locator_type = locator_type
        self.__is_collection = False
        self.index = 0
        # TODO: Investigate if should be used in functionality or should be
        # removed.
        # pylint: disable=unused-private-member
        self.__errors = []
        self.__is_not = False

    def __put_log(self, message):
        if self.logger is not None:
            if self.logger == "behave":
                logger.log(f"{message} \n")
            else:
                logger.log(message)

    def __show_error(self, exception):
        testui_error(self.testui_driver, exception)

    def get(self, index):
        self.__is_collection = True
        self.index = index

        return self

    def get_element(self, index=0) -> WebElement:
        if self.__is_collection:
            return self.__find_by_collection()[self.index]
        if index != 0:
            return self.__find_by_collection()[index]

        return self.__find_by_element()

    def __find_by_element(self) -> WebElement:
        if self.locator_type == "id":
            return self.driver.find_element(by=By.ID, value=self.locator)

        if self.locator_type == "css":
            return self.driver.find_element(
                by=By.CSS_SELECTOR, value=self.locator
            )

        if self.locator_type == "className":
            return self.driver.find_element(
                by=By.CLASS_NAME, value=self.locator
            )

        if self.locator_type == "name":
            return self.driver.find_element(By.NAME, self.locator)

        if self.locator_type == "xpath":
            return self.driver.find_element(By.XPATH, self.locator)

        if self.locator_type == "android_id_match":
            return self.driver.find_element_by_android_uiautomator(
                f'resourceIdMatches("{self.locator}")'
            )

        if self.locator_type == "accessibility":
            return self.driver.find_element_by_accessibility_id(self.locator)

        if self.locator_type == "uiautomator":
            return self.driver.find_element_by_android_uiautomator(self.locator)

        if self.locator_type == "classChain":
            return self.driver.find_element_by_ios_class_chain(self.locator)

        if self.locator_type == "predicate":
            return self.driver.find_element_by_ios_predicate(self.locator)

        raise ElementException(f"locator not supported: {self.locator_type}")

    def __find_by_collection(self) -> List[WebElement]:
        if self.locator_type == "id":
            return self.driver.find_elements(by=By.ID, value=self.locator)

        if self.locator_type == "css":
            return self.driver.find_elements(
                by=By.CSS_SELECTOR, value=self.locator
            )

        if self.locator_type == "className":
            return self.driver.find_elements(
                by=By.CLASS_NAME, value=self.locator
            )

        if self.locator_type == "name":
            return self.driver.find_elements(By.NAME, self.locator)

        if self.locator_type == "xpath":
            return self.driver.find_elements(By.XPATH, self.locator)

        if self.locator_type == "android_id_match":
            return self.driver.find_elements_by_android_uiautomator(
                f'resourceIdMatches("{self.locator}")'
            )

        if self.locator_type == "accessibility":
            return self.driver.find_elements_by_accessibility_id(self.locator)

        if self.locator_type == "uiautomator":
            return self.driver.find_elements_by_android_uiautomator(
                self.locator
            )

        if self.locator_type == "classChain":
            return self.driver.find_elements_by_ios_class_chain(self.locator)

        if self.locator_type == "predicate":
            return self.driver.find_elements_by_ios_predicate(self.locator)

        raise ElementException(f"locator not supported: {self.locator_type}")

    def is_visible(self, log=True, **kwargs) -> bool:
        is_not = False

        # Allows passing "is_not" as a kwarg to not overwrite self.__is_not.
        # This is not meant to be used by the user.
        if "is_not" in kwargs:
            is_not = kwargs["is_not"]
        else:
            is_not = self.__is_not
            self.__is_not = False

        is_visible = False
        try:
            is_visible = self.get_element().is_displayed()

            if log:
                if is_visible:
                    self.__put_log(
                        f'{self.device_name}: element "{self.locator_type}: '
                        f'{self.locator}" is visible'
                    )
                else:
                    self.__put_log(
                        f'{self.device_name}: element "{self.locator_type}: '
                        f'{self.locator}" is not visible'
                    )
        except Exception:
            if log:
                self.__put_log(
                    f'{self.device_name}: element "{self.locator_type}: '
                    f'{self.locator}" is not visible'
                )

        if is_not:
            return not is_visible
        else:
            return is_visible

    def is_visible_in(self, seconds):
        start = time.time()
        is_not = self.__is_not
        while time.time() < start + seconds:
            self.__is_not = is_not
            if self.is_visible(log=False):
                return True
        return False

    def visible_for(self, seconds=1):
        start = time.time()
        is_not = self.__is_not
        err_text = "not"
        if self.__is_not:
            err_text = "is"
        while time.time() < start + seconds:
            self.__is_not = is_not
            if not self.is_visible():
                return self.__show_error(
                    f"{logger.bcolors.FAIL} {self.device_name} Element "
                    f"{err_text} found with the following locator: "
                    f'"{self.locator_type}:{self.locator}" during the time of '
                    f"{time.time() - start}s {logger.bcolors.ENDC}"
                )
        if is_not:
            self.__put_log(
                f'{self.device_name}: element "{self.locator_type}: '
                f'{self.locator}" is not visible for {seconds}s'
            )
        else:
            self.__put_log(
                f'{self.device_name}: element "{self.locator_type}: '
                f'{self.locator}" is visible for {seconds}s'
            )
        return self

    def wait_until_visible(self, seconds=10.0, log=True) -> "Elements":
        start = time.time()

        is_not = self.__is_not
        self.__is_not = False

        is_visible = self.is_visible(log=False, is_not=is_not)
        while time.time() < start + seconds and not is_visible:
            time.sleep(0.2)
            is_visible = self.is_visible(log=False, is_not=is_not)

        if is_visible:
            if log:
                self.__put_log(
                    f'{self.device_name}: Element "{self.locator_type}: '
                    f'{self.locator}" was visible in {time.time() - start}s'
                )

            return self

        err_text = "was not"
        if is_not:
            err_text = "was"

        self.__show_error(
            f"{self.device_name}: Element {err_text} found with the following "
            f'locator: "{self.locator_type}: {self.locator}" '
            f"after {time.time() - start}s"
        )

        return self

    def wait_until_attribute(self, attr, text, seconds=10):
        start = time.time()
        err = None
        value = ""
        err_text, info_text = "should have been", "="
        if self.__is_not:
            err_text, info_text = "should not have been", "!="
        while time.time() < start + seconds:
            try:
                value = self.get_element().get_attribute(attr)
                if (value == text) != self.__is_not:
                    self.__put_log(
                        f'{self.device_name}: element "{self.locator_type}: '
                        f'{self.locator}" has attribute "{attr}" -> "{value}" '
                        f'{info_text} "{text}" after {time.time() - start}s'
                    )
                    self.__is_not = False
                    return self
            except Exception as error:
                err = error
                time.sleep(0.2)
        if err is None:
            err = ""
        return self.__show_error(
            f"{logger.bcolors.FAIL}{err} {self.device_name} Element {err_text} "
            f"found with the following locator: "
            f'"{self.locator_type}:{self.locator}" and '
            f'attribute "{attr}" -> "{value}" {err_text} "{text}" after '
            f"{time.time() - start}s{logger.bcolors.ENDC}"
        )

    def wait_until_contains_attribute(self, attr, text, seconds=10):
        start = time.time()
        err = None
        value = ""
        info_text, err_text = "containing", " not"
        if self.__is_not:
            info_text, err_text = "not containing", " "
        while time.time() < start + seconds:
            try:
                value = self.get_element().get_attribute(attr)
                if value.__contains__(text) != self.__is_not:
                    self.__put_log(
                        f'element "{self.locator_type}: {self.locator}" has '
                        f'attribute "{attr}" {info_text} "{text}" after '
                        f"{time.time() - start}s"
                    )
                    self.__is_not = False
                    return self
            except Exception as error:
                err = error
                time.sleep(0.2)
        if err is None:
            err = ""
        return self.__show_error(
            f"{logger.bcolors.FAIL}{err} {self.device_name} Element{err_text} "
            f"found with the following locator: "
            f'"{self.locator_type}:{self.locator}" and '
            f'attribute "{attr}" -> "{value}" {info_text} "{text}" '
            f"after {time.time() - start}s {logger.bcolors.ENDC}"
        )

    def wait_until_contains_sensitive_attribute(
        self, attr, text, seconds=10.0, log=True
    ):
        start = time.time()
        err = None
        value = ""
        info_text, err_text = "contains", "not"
        if self.__is_not:
            info_text, err_text = "not contains", "is"
        while time.time() < start + seconds:
            try:
                value = self.get_element().get_attribute(attr)
                if value.lower().__contains__(text.lower()) != self.__is_not:
                    self.__put_log(
                        f'{self.device_name}: element "{self.locator_type}: '
                        f'{self.locator}" has attribute "{attr}" -> "{value}" '
                        f'{info_text} no case sensitive value "{text}" after '
                        f"{time.time() - start}s"
                    )
                    self.__is_not = False
                    return self
            except Exception as error:
                err = error
                time.sleep(0.2)
        if err is None:
            err = ""
        if log:
            return self.__show_error(
                f"{logger.bcolors.FAIL}{err} {self.device_name} Element "
                f"{err_text} found with the following locator: "
                f'"{self.locator_type}:{self.locator}" and attribute "{attr}" '
                f'containing no case sensitive value "{text}" {err_text} '
                f'"{value}" after {time.time() - start}s {logger.bcolors.ENDC}'
            )

        raise ElementException(err)

    def no(self, is_not=True):
        self.__is_not = is_not
        return self

    def click(self):
        timeout = 5  # [seconds]
        start = time.time()

        err = None
        while time.time() < start + timeout:
            try:
                element = self.get_element()
                element.click()
                self.__put_log(
                    f'{self.device_name}: element "{self.locator_type}: '
                    f'{self.locator}" pressed for {time.time() - start}s'
                )
                return self
            except Exception as error:
                err = error
                time.sleep(0.2)
        return self.__show_error(
            f"{logger.bcolors.FAIL}{err} {self.device_name}: Element "
            f'"{self.locator_type}: {self.locator}" could not be clicked after '
            f"{time.time() - start}s {logger.bcolors.ENDC}"
        )

    def press_hold_for(self, milliseconds=1000):
        timeout = 5  # [seconds]
        start = time.time()

        err = None
        while time.time() < start + timeout:
            try:
                self.get_element()
                try:
                    is_browser: str = self.testui_driver.context
                    if is_browser.__contains__("NATIVE"):
                        browser = False
                    else:
                        browser = True
                except Exception:
                    browser = True
                if not browser:
                    ta = TouchAction(self.driver)
                    ta.press(self.get_element()).wait(
                        milliseconds
                    ).release().perform()
                else:
                    ta = ActionChains(self.driver)
                    ta.click_and_hold(self.get_element()).pause(
                        milliseconds // 1000
                    ).release().perform()

                self.__put_log(
                    f'{self.device_name}: element "{self.locator_type}: '
                    f'{self.locator}" pressed for {time.time() - start}s'
                )
                return self
            except Exception as error:
                err = error
                time.sleep(0.2)
        return self.__show_error(
            f"{logger.bcolors.FAIL}{err} {self.device_name}: Element not found "
            f'with the following locator: "{self.locator_type}:{self.locator}" '
            f"after {time.time() - start}s {logger.bcolors.ENDC}"
        )

    def click_by_coordinates(self, x, y):
        timeout = 5  # [seconds]
        start = time.time()

        err = None
        while time.time() < start + timeout:
            try:
                self.get_element()
                time.sleep(1)
                ta = TouchAction(self.driver)
                ta.tap(x=x, y=y).perform()
                self.__put_log(
                    f'{self.device_name}: element "x={x}: y={y}" clicked after '
                    f"{time.time() - start}s"
                )
                return self
            except Exception as error:
                err = error
                time.sleep(0.2)
        return self.__show_error(
            f"{logger.bcolors.FAIL}{err} {self.device_name}: Element not found "
            f'with the following locator: "{self.locator_type}:{self.locator}" '
            f"after {time.time() - start}s {logger.bcolors.ENDC}"
        )

    @property
    def location(self):
        return Dimensions(
            self.get_element().location.get("x"),
            self.get_element().location.get("y"),
        )

    @property
    def dimensions(self):
        return Dimensions(
            self.get_element().size.get("width"),
            self.get_element().size.get("height"),
        )

    def screenshot(self, image_name="cropped_image.png"):
        """
        Takes screenshot of the specific element
        :param image_name: relative path to image taken
        :return:
        """
        self.wait_until_visible()
        self.testui_driver.save_screenshot(f"{self.device_name}-crop_image.png")
        dimensions = self.dimensions
        top_left = self.location
        ImageRecognition(
            f"testui-{self.device_name}-crop_image.png"
        ).crop_original_image(
            top_left.x + dimensions.x // 2,
            top_left.y + dimensions.y // 2,
            dimensions.x,
            dimensions.y,
            image_name,
        )

        root_dir = self.testui_driver.configuration.screenshot_path
        if not root_dir:
            root_dir = path.dirname(
                path.dirname(path.dirname(path.abspath(__file__)))
            )

        os.remove(root_dir + f"/testui-{self.device_name}-crop_image.png")

        return self

    def find_image_match(
        self,
        image_name,
        threshold=0.9,
        image_match="",
        max_scale=2.0,
        min_scale=0.3,
    ):
        """
        Takes screenshot of the element and compares it with the one you provide
        as 'image_name'
        :param min_scale:
        :param max_scale:
        :param image_match: returns the image with a rectangle showing the match
        :param image_name: relative path to image
        :param threshold: limit to consider image as a match (0 to 1)
        :return: Elements
        """
        is_not = self.__is_not
        self.__is_not = False
        self.screenshot(self.device_name + ".png")
        found, precision = ImageRecognition(
            self.device_name + ".png", image_name, threshold, self.device_name
        ).compare(
            image_match=image_match, max_scale=max_scale, min_scale=min_scale
        )
        if not found and not is_not:
            self.__show_error(
                f"{self.device_name}: The images compared did not match. "
                f"Threshold={threshold}, matched = {precision}"
            )
        if found and is_not:
            self.__is_not = False
            self.__show_error(
                f"{self.device_name}: The images compared matched. "
                f"Threshold={threshold}, matched = {precision}"
            )
        root_dir = (
            os.path.dirname(
                os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
            )
            + "/"
        )
        os.remove(root_dir + self.device_name + ".png")

        return self

    def is_image_match(
        self,
        image_name,
        threshold=0.9,
        image_match="",
        max_scale=2.0,
        min_scale=0.3,
    ):
        """
        Takes screenshot of the element and compares it with the one you provide
        as 'image_name'
        :param min_scale:
        :param max_scale:
        :param image_match: returns the image with a rectangle showing the match
        :param image_name: relative path to image
        :param threshold: limit to consider image as a match (0 to 1)
        :return: Elements
        """
        is_not = self.__is_not
        self.__is_not = False
        self.screenshot(self.device_name + ".png")
        found, _ = ImageRecognition(
            self.device_name + ".png", image_name, threshold, self.device_name
        ).compare(
            image_match=image_match, max_scale=max_scale, min_scale=min_scale
        )
        if not found and not is_not:
            return False
        if found and is_not:
            return False
        root_dir = (
            os.path.dirname(
                os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
            )
            + "/"
        )
        os.remove(root_dir + self.device_name + ".png")

        return True

    def swipe(
        self,
        start_x=None,
        start_y=None,
        end_x=None,
        end_y=None,
        el=None,
        duration=None,
    ):
        """
        It swipes from element to the el(second element) or to the coordinates
        you define with start_x, start_y, end_x, end_y. end_x and end_y are
        mandatory if el(second element) is not defined.
        :param start_x: Coordinate x from where it will begin swiping
        :param start_y: Coordinate y from where it will begin swiping
        :param end_x: Coordinate x from where it will end swiping
        :param end_y: Coordinate y from where it will end swiping
        :param el: element 'Elements' type from where it will end swiping
        (instead of end_x, end_y)
        :param duration: duration of the swiping action
        :return: Elements
        """
        timeout = 5  # [seconds]
        start = time.time()

        err = None
        while time.time() < start + timeout:
            try:
                location = self.location
                if start_x is None:
                    start_x = location.x
                if start_y is None:
                    start_y = location.y
                time.sleep(0.1)
                el: Elements
                if el is not None:
                    location2 = el.location
                    if end_x is None:
                        end_x = location2.x
                    if end_y is None:
                        end_y = location2.y
                    action = TouchAction(self.driver)
                    action.press(x=start_x, y=start_y).wait(duration).move_to(
                        x=end_x, y=end_y
                    ).release()
                    action.perform()
                else:
                    if end_x is None:
                        end_x = location.x
                    if end_y is None:
                        end_y = location.y
                    self.driver.swipe(
                        start_y=start_y,
                        start_x=start_x,
                        end_y=end_y,
                        end_x=end_x,
                        duration=duration,
                    )
                return self
            except Exception as error:
                err = error
        return self.__show_error(
            f"{logger.bcolors.FAIL}{err} {self.device_name}: Element not found "
            f'with the following locator: "{self.locator_type}:{self.locator}" '
            f"after {time.time() - start}s {logger.bcolors.ENDC}"
        )

    def slide_percentage(self, percentage, start_x=None):
        width = self.dimensions.x * percentage / 100
        end_width = width + self.location.x
        if start_x is None:
            start_x = self.location.x
        self.swipe(start_x=start_x, end_x=end_width)

    def swipe_until_text(
        self,
        start_x=None,
        start_y=None,
        end_x=None,
        end_y=None,
        text=None,
        el=None,
        max_swipes=50,
    ):
        """
        Swipe until an element with that text appears and Returns the element
        that contains the text.
        endX and endY are mandatory if el(element) is not defined.
        text is mandatory
        :param start_x:
        :param start_y:
        :param end_x:
        :param end_y:
        :param text:
        :param el:
        :param max_swipes:
        :return: Elements
        """

        if el is None and (start_x is None or end_y is None):
            raise Exception(
                "if element not specified, end_x and end_y are required "
                "to be able to swipe"
            )
        if text is None:
            raise Exception("text cannot be None for swipe_until_text method")
        start = time.time()
        for _ in range(max_swipes):
            try:
                e(
                    self.testui_driver, "uiautomator", f'textContains("{text}")'
                ).wait_until_visible(0.1, False)
                break
            except Exception:
                self.swipe(start_x, start_y, end_x, end_y, el, None)
        self.__put_log(
            f'{self.device_name}: element "{self.locator_type}: '
            f'{self.locator}" with text {text} '
            f"found after {time.time() - start}s"
        )
        return e(self.testui_driver, "uiautomator", f'textContains("{text}")')

    def send_keys(self, value, log=True):
        timeout = 10  # [seconds]
        start = time.time()
        if value is None:
            raise Exception("keys sent cannot be None type")
        err = None
        while time.time() < start + timeout:
            try:
                self.get_element().send_keys(value)
                if log:
                    self.__put_log(
                        f'{self.device_name}: element "{self.locator_type}: '
                        '{self.locator}" received keys "{value}" after '
                        f"{time.time() - start}s"
                    )
                return self
            except Exception as error:
                err = error
                time.sleep(0.2)
        return self.__show_error(
            f"{logger.bcolors.FAIL}{err} {self.device_name}: Element not found "
            f'with the following locator: "{self.locator_type}:{self.locator}" '
            f"after {time.time() - start}s {logger.bcolors.ENDC}"
        )

    def clear(self) -> "Elements":
        self.get_element().clear()

        return self

    def get_text(self):
        timeout = 10  # [seconds]
        start = time.time()

        err = None
        while time.time() < start + timeout:
            try:
                text = self.get_element().text
                self.__put_log(
                    f'{self.device_name}: element "{self.locator_type}: '
                    f'{self.locator}" has text "{text}" {time.time() - start}s'
                )
                return text
            except Exception as error:
                err = error
                time.sleep(0.2)
        self.__show_error(
            f"{logger.bcolors.FAIL}{err} {self.device_name}: Element not found "
            f'with the following locator: "{self.locator_type}:{self.locator}" '
            f"after {time.time() - start}s {logger.bcolors.ENDC}"
        )

    def get_value(self):
        timeout = 10  # [seconds]
        start = time.time()

        err = None
        while time.time() < start + timeout:
            try:
                text = self.get_element().get_attribute("value")
                self.__put_log(
                    f'{self.device_name}: element "{self.locator_type}: '
                    f'{self.locator}" has text "{text}" {time.time() - start}s'
                )
                return text
            except Exception as error:
                err = error
                time.sleep(0.2)
        self.__show_error(
            f"{logger.bcolors.FAIL}{err} {self.device_name}: Element not found "
            f'with the following locator: "{self.locator_type}:{self.locator}" '
            f"after {time.time() - start}s {logger.bcolors.ENDC}"
        )

    def get_name(self):
        timeout = 10  # [seconds]
        start = time.time()

        err = None
        while time.time() < start + timeout:
            try:
                text = self.get_element().get_attribute("name")
                self.__put_log(
                    f'{self.device_name}: element "{self.locator_type}: '
                    f'{self.locator}" has name "{text}" {time.time() - start}s'
                )
                return text
            except Exception as error:
                err = error
                time.sleep(0.2)
        self.__show_error(
            f"{logger.bcolors.FAIL}{err} {self.device_name}: Element not "
            "found with the following locator: "
            f'"{self.locator_type}:{self.locator}" after '
            f"{time.time() - start}s {logger.bcolors.ENDC}"
        )

    def get_attribute(self, att):
        timeout = 10  # [seconds]
        start = time.time()

        err = None
        while time.time() < start + timeout:
            try:
                text = self.get_element().get_attribute(att)
                self.__put_log(
                    f'{self.device_name}: element "{self.locator_type}: '
                    f'{self.locator}" has "{att}: {text}" '
                    f"{time.time() - start}s"
                )
                return text
            except Exception as error:
                err = error
                time.sleep(0.2)
        self.__show_error(
            f"{logger.bcolors.FAIL}{err} {self.device_name}: Element not found "
            f'with the following locator: "{self.locator_type}:{self.locator}" '
            f"after {time.time() - start}s {logger.bcolors.ENDC}"
        )

    def press_and_compare(
        self,
        image,
        milliseconds=1000,
        threshold=0.9,
        fps_reduction=1,
        keep_image_as="",
    ):
        self.testui_driver.start_recording_screen()
        self.press_hold_for(milliseconds)

        found = ""
        not_found = "not"
        if self.__is_not:
            found = "not"
            not_found = ""

        start = time.time()

        if self.testui_driver.stop_recording_and_compare(
            image, threshold, fps_reduction, self.__is_not, keep_image_as, False
        ):
            self.__put_log(
                f"{self.device_name}: image {found} found while pressing "
                f'element "{self.locator_type}: {self.locator}" '
                f"after {time.time() - start}s"
            )
        else:
            self.__show_error(
                f"{self.device_name}: image {not_found} found while pressing "
                f'element "{self.locator_type}: {self.locator}" '
                f"after {time.time() - start}s"
            )
        self.__is_not = False
        return self

    def collection_size(self):
        self.__is_collection = False
        index = self.index
        self.index = 0
        self.wait_until_visible()
        self.index = index
        self.__is_collection = True
        return len(self.__find_by_collection())

    def find_by_attribute(
        self, attribute, value: str, timeout=10, case_sensitive=True
    ):
        start = time.time()
        self.wait_until_visible()
        self.__is_collection = True
        while time.time() < start + timeout:
            for i, element in enumerate(self.__find_by_collection()):
                if case_sensitive and element.get_attribute(attribute) == value:
                    self.__put_log(
                        f"{self.device_name}: element in collection "
                        f'{self.locator_type}: {self.locator}" with attribute '
                        f'"{attribute}" = "{value}" found after '
                        f"{time.time() - start}s"
                    )
                    self.index = i
                    return self
                if (
                    not case_sensitive
                    and element.get_attribute(attribute).lower()
                    == value.lower()
                ):
                    self.__put_log(
                        f"{self.device_name}: element in collection "
                        f'{self.locator_type}: {self.locator}" with attribute '
                        f'"{attribute}" = "{value}" found after '
                        f"{time.time() - start}s"
                    )
                    self.index = i
                    return self
        self.__show_error(
            f"{self.device_name}: no element in collection "
            f'"{self.locator_type}: {self.locator}" had attribute '
            f'"{attribute}" = "{value}" after {time.time() - start}s'
        )


def e(driver, locator_type: str, locator: str) -> Elements:
    """
    Args:
        driver: Driver that is going to fetch element.
        locator_type (str): Type of the pattern in which locator is written.
        locator (str): Pattern which describes target.

    Possible locator types:
        id,
        css,
        className,
        name,
        xpath,
        accessibility,
        uiautomator,
        classChain,
        predicate
    """

    # TODO: Locator types should be enums.

    return Elements(driver, locator_type, locator)


def scroll_by_text(driver, text, element=None, exact_text=False) -> Elements:
    if exact_text:
        method_text = "text"
    else:
        method_text = "textContains"
    if element is None:
        element = "scrollable(true)"

    return e(
        driver,
        "uiautomator",
        (
            f"new UiScrollable(new UiSelector().{element})."
            f'scrollIntoView({method_text}("{text}"));'
        ),
    )


def scroll_by_resource_id(driver, res_id) -> Elements:
    locator = (
        "new UiScrollable(new UiSelector().scrollable(true)).scrollIntoView"
        f'(new UiSelector().resourceId("{res_id}"));'
    )

    return e(driver, "uiautomator", locator)


class AndroidLocator:
    @classmethod
    def scroll(cls, method: str, scrollable_element=None):
        if scrollable_element is None:
            scrollable_element = "scrollable(true)"

        return (
            f"new UiScrollable(new UiSelector()."
            f"{scrollable_element}).scrollIntoView({method});"
        )

    @classmethod
    def text(cls, text: str):
        return f'text("{text}")'

    @classmethod
    def text_contains(cls, text: str):
        return f'textContains("{text}")'

    @classmethod
    def id_match(cls, text: str):
        return f'resourceIdMatches("{text}")'

    @classmethod
    def class_name(cls, text: str):
        return f'className("{text}")'

    @classmethod
    def parent(cls, parent_method, child_method):
        return f"fromParent({parent_method}).{child_method}"

    @classmethod
    def child(cls, parent_method, child_method):
        return f"childSelector({child_method}).{parent_method}"



"""Py-TestUI setup module"""

import pathlib
from setuptools import find_packages, setup


root_path = pathlib.Path(__file__).parent.resolve()
long_description = (root_path / "README.md").read_text(encoding="utf-8")


setup(
    name="python-testui",
    version="1.1.1",
    description="Browser and Mobile automation framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/testdevlab/Py-TestUI",
    author="Alvaro Santos Laserna Lopez",
    license="Apache LICENSE 2.0",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Quality Assurance",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    keywords="Py-TestUI,mobile automation,browser automation",
    project_urls={
        "Source": "https://github.com/testdevlab/Py-TestUI",
        "Tracker": "https://github.com/testdevlab/Py-TestUI/issues",
        "TestDevLab": "https://www.testdevlab.com/",
    },
    packages=find_packages(),
    python_requires=">=3.6, <4",
    install_requires=[
        "pytest==6.2.5",
        "Appium-Python-Client==2.6.0",
        "selenium==4.1.0",
        "opencv-python==4.6.0.66",
        "geckodriver-autoinstaller==0.1.0",
        "value==0.1.0",
        "pytest-xdist==2.5.0",
        "pytest-testrail==2.9.0",
        "pure-python-adb==0.3.0.dev0",
        "webdriver-manager==3.6.3",
        "numpy==1.22.0",
        "imutils==0.5.4",
    ],
)


# #r2DZTbXrhLc



import argparse
import multiprocessing as mp
import os
import time
import subprocess

from testui.support import logger


def parallel_testui():
    remove_logs()
    args = __arg_parser()
    try:
        os.remove("report_fails.txt")
    except FileNotFoundError:
        pass
    try:
        os.remove("report_cases.txt")
    except FileNotFoundError:
        pass
    test_run_id = None
    if args.testrail_id is not None:
        test_run_id = args.testrail_id
        logger.log_info(f"Starting testrail run {test_run_id}")
    elif args.testrail is not None:
        test_run_id = __start_run_id(args, args.testrail)
        logger.log_info(
            f"Created testrail run {test_run_id} "
            f"with testrail name: {args.testrail}"
        )
    number_of_cases = get_total_number_of_cases(args)
    logger.log_info(
        f"----------- Total Number of Test Cases: {number_of_cases} -----------"
    )
    start = time.time()
    __start_processes(args.markers, args, test_run_id)
    end_time = time.time() - start
    f = open("fails.txt")
    fails = f.read()

    passed = 0
    fail = 0
    for letter in fails:
        if letter == "0":
            passed += 1
        if letter == "2":
            fail += 1
    if fail != 0:
        os.remove("fails.txt")
        number_of_fails = __check_number_of_fails()
        percentage = 100 - 100 * number_of_fails / number_of_cases
        logger.log_error(
            f"----------- Total Number of Test Cases: {number_of_cases}. "
            f"Percentage Passed {percentage.__round__(3)}% -----------"
        )
        logger.log_error(
            f"----------- Tests Finished with errors! "
            f"{fail} markers failed and {passed} passed."
            f" Execution Time {__seconds_to_minutes(end_time)} -----------"
        )
        logger.log_error(
            "----------- Total Number of Test Failed: "
            f"{__check_number_of_fails()} -----------"
        )
        logger.log_error(f"There were test fails: \n\n{__check_txt_fails()}")
        raise Exception("----------- Tests Finished with errors! -----------")

    try:
        os.remove("fails.txt")
    except FileNotFoundError:
        logger.log_error(
            "There were not errors file, there might be some other issues "
            "during the execution..."
        )
    logger.log_pass(
        f"----------- Total Number of Test Cases: {number_of_cases} -----------"
    )
    logger.log_pass(
        f"----------- Tests Finished successfully. {passed} markers passed. "
        f"Total Execution Time {__seconds_to_minutes(end_time)} -----------"
    )


def remove_logs():
    root_dir = os.path.dirname(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    )
    i = 0
    try:
        for filename in os.listdir(root_dir + "/report_screenshots"):
            file_path = os.path.join(
                root_dir + "/report_screenshots/", filename
            )
            os.remove(file_path)
            if i == 0:
                logger.log("Cleaning report_screenshots folder...")
                i += 1
    except FileNotFoundError:
        pass
    i = 0
    try:
        for filename in os.listdir(root_dir + "/appium_logs"):
            file_path = os.path.join(root_dir + "/appium_logs/", filename)
            os.remove(file_path)
            if i == 0:
                logger.log("Cleaning appium_logs folder...")
                i += 1
    except FileNotFoundError:
        pass


def get_total_number_of_cases(args):
    number_of_cases = 0
    for marker in args.markers:
        output = subprocess.run(
            [
                "pytest",
                "-m",
                f"{marker} {args.general_markers}",
                "--collect-only",
            ],
            stdout=subprocess.PIPE,
            check=False,
        )
        response = output.stdout
        string_response = response.__str__()
        if string_response.__contains__(" / "):
            for cases in string_response.split(" / "):
                if cases.__contains__(" selected"):
                    number = cases.split(" selected")[0]
                    number_of_cases += int(number)
    if args.single_thread_marker is not None:
        output = subprocess.run(
            [
                "pytest",
                "-m",
                f"{args.single_thread_marker} {args.general_markers}",
                "--collect-only",
            ],
            stdout=subprocess.PIPE,
            check=False,
        )
        response = output.stdout
        string_response = response.__str__()
        if string_response.__contains__(" / "):
            for cases in string_response.split(" / "):
                if cases.__contains__(" selected"):
                    number = cases.split(" selected")[0]
                    number_of_cases += int(number)
    return number_of_cases


def __seconds_to_minutes(time_seconds):
    minutes = int(time_seconds // 60)
    seconds = int(time_seconds % 60)
    if minutes < 10:
        ms = f"0{minutes}"
    else:
        ms = f"{minutes}"
    if seconds < 10:
        s = f"0{seconds}"
    else:
        s = f"{seconds}"
    return f"{ms}:{s} ({time_seconds}s)"


def __arg_parser():
    parser = argparse.ArgumentParser(description="Parallel Testing")
    parser.add_argument(
        "markers",
        metavar="N",
        type=str,
        nargs="+",
        help="an string for the accumulator",
    )

    parser.add_argument(
        "--parallel", type=int, help="number of parallel threads"
    )

    parser.add_argument(
        "--s",
        type=__str2bool,
        nargs="?",
        const=True,
        default=False,
        help="make more verbose logs (pytest -s)",
    )

    parser.add_argument(
        "--testrail",
        type=str,
        help="Use test rail and specify name. "
        "You will have to add a file testrail.cfg in root project dir",
    )

    parser.add_argument(
        "--testrail_id",
        type=str,
        help="Use test rail and specify testrail id. "
        "You will have to add a file testrail.cfg in root project dir",
    )

    parser.add_argument(
        "--testrail_pwd",
        type=str,
        help="Use test rail and specify password. "
        "You will have to add a file testrail.cfg in root project dir",
    )

    parser.add_argument(
        "--general", type=str, help="general flags to put in all threads"
    )

    parser.add_argument(
        "--general_markers", type=str, help="markers included in all tests"
    )

    parser.add_argument(
        "--single_thread_marker",
        type=str,
        help="marker to run without any thread running in parallel",
    )

    parser.add_argument(
        "--udids", nargs="*", type=str, default=[], help="device udids"
    )

    args = parser.parse_args()
    logger.log(f"Test Run for {args.markers}")
    if args.general_markers is None:
        args.general_markers = ""
    else:
        logger.log(f"General marker added: {args.general_markers}")
    if args.udids:
        logger.log(f"Devices to use for automation: {args.udids}")
        for i, udid in enumerate(args.udids):
            os.environ[f"UDID{i}"] = udid
    if args.parallel is None:
        args.parallel = 1
    if args.parallel > len(args.markers):
        args.parallel = len(args.markers)
    logger.log(f"Number of parallel threads: {args.parallel}")

    if args.general is None:
        args.general = ""
    else:
        logger.log(f"General flags: {args.general}")
    if args.single_thread_marker is not None:
        logger.log_info(
            "Single Thread marker: "
            f"{args.single_thread_marker} {args.general_markers}"
        )
    return args


def __str2bool(v):
    if isinstance(v, bool):
        return v

    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    if v.lower() in ("no", "false", "f", "n", "0"):
        return False

    raise argparse.ArgumentTypeError("Boolean value expected.")


def __start_processes(markers: list, args, test_run_id=None):
    ps = []
    amount = len(markers) // args.parallel
    amount_plus = 0
    if len(markers) % args.parallel != 0:
        amount_plus = len(markers) % args.parallel
    i = 0
    for j in range(args.parallel):
        tags = []
        if j != args.parallel - 1:
            for _ in range(amount):
                if len(markers) > i:
                    tags.append(markers[i])
                    i += 1
        else:
            for _ in range(amount + amount_plus):
                if len(markers) > i:
                    tags.append(markers[i])
                    i += 1
        logger.log_info(f"Thread {j} has markers: {tags}")
        ps.append(
            mp.Process(target=__process, args=(tags, args, j, test_run_id))
        )
    for p in ps:
        p.daemon = True
        p.start()
    for p in ps:
        p.join()
    if args.single_thread_marker is not None:
        __process([args.single_thread_marker], args, 0, test_run_id)


def __start_run_id(args, test_run_name):
    end_marker = ""
    for i, marker in enumerate(args.markers):
        if i == len(args.markers):
            end_marker += f"{marker} {args.general_markers}"
        else:
            end_marker += f"{marker} {args.general_markers} or "
    if args.single_thread_marker is not None:
        end_marker += f"{args.single_thread_marker} {args.general_markers}"
    with open("testrail_id_file.txt", "wb") as out:
        cmd = [
            "pytest",
            "-m",
            end_marker,
            "--testrail",
            "--tr-config=testrail.cfg",
            f"--tr-testrun-name={test_run_name}",
        ]
        if args.testrail_pwd is not None:
            cmd.append(f"--tr-password={args.testrail_pwd}")
        process = subprocess.Popen(cmd, stdout=out, stderr=out)
    i = 0
    while True:
        time.sleep(0.1)
        out = open("testrail_id_file.txt")
        text = out.read()
        if text.__contains__("New testrun created") or i > 50:
            out.close()
            process.terminate()
            process.wait()
            os.remove("testrail_id_file.txt")
            if text.__contains__("ID="):
                id_test = text.split("ID=")[1]
                if text.split("ID=")[1].__contains__("\n"):
                    id_test = text.split("ID=")[1].split("\n")[0]
                logger.log_info(f"Test run: {id_test}")
                return id_test
            raise Exception("Failed to create Test Run")
        if text.__contains__("Failed to create testrun"):
            out.close()
            process.send_signal(signal=2)
            process.terminate()
            process.wait()
            os.remove("testrail_id_file.txt")
            logger.log_error(f"Failed to create Test Run: \n {text}")
            raise Exception("Failed to create Test Run")
        out.close()
        i += 1


def __process(markers: list, args, thread=0, test_run_id=None):
    for marker in markers:
        try:
            os.remove(f".my_cache_dir_{thread}/v/cache/lastfailed")
            os.remove(f".my_cache_dir_{thread}/v/cache/nodeids")
            os.remove(f".my_cache_dir_{thread}/v/cache/stepwise")
        except FileNotFoundError:
            pass
        quiet = "-q"
        if args.s:
            quiet = "-s"
        testrail = ""
        if test_run_id is not None:
            testrail = (
                f"--testrail --tr-config=testrail.cfg --tr-run-id={test_run_id}"
            )
            if args.testrail_pwd is not None:
                testrail += f" --tr-password={args.testrail_pwd}"
        cache = f"-o cache_dir=.my_cache_dir_{thread}"
        start_1 = time.time()
        logger.log(
            f'Starting: pytest {quiet} -m "{marker} {args.general_markers}" '
            f"{testrail} {args.general} {cache}"
        )
        pr_1 = os.system(
            f'pytest {quiet} -m "{marker} {args.general_markers}" '
            f"{testrail} {args.general} {cache}"
        )
        file = open("fails.txt", "a+")
        file.write(f"{pr_1}")
        file.close()
        if f"{pr_1}" == "0":
            logger.log_pass(
                f'FINISHED RUN WITH MARKERS: "{marker} {args.general_markers}" '
                f"AFTER {__seconds_to_minutes(time.time() - start_1)}"
            )
        else:
            logger.log_error(
                "FINISHED RUN WITH ERRORS AFTER "
                f"{__seconds_to_minutes(time.time() - start_1)}. "
                f'MARKERS: "{marker} {args.general_markers}"'
            )
        __set_fails_file(f".my_cache_dir_{thread}")


def __set_fails_file(cache_dir):
    fails = __check_fails(cache_dir)
    for fail in fails:
        file = open("report_fails.txt", "a+")
        file.write(f"{fail} \n")
        file.close()


def __check_number_of_fails():
    try:
        file = open("report_fails.txt")
        fails = file.read()
        file.close()
        return len(fails.split("tests/")) - 1
    except FileNotFoundError:
        return 0


def __check_txt_fails():
    try:
        file = open("report_fails.txt")
        text = file.read()
        file.close()
        return text
    except FileNotFoundError:
        return ""


def __check_fails(cache=".pytest_cache"):
    try:
        f = open(f"{cache}/v/cache/lastfailed")
        fails = f.read()
        fails_tests = []
        i = 0
        for fail in fails.split("true"):
            if i == 0:
                fails_tests.append(__clean_str(fail))
            elif i == len(fails) - 1:
                f.close()
                os.remove(f"{cache}")
                return fails_tests
            else:
                fails_tests.append(__clean_str(fail))
            i += 1
        return fails_tests
    except FileNotFoundError:
        return []


def __clean_str(string: str):
    return (
        string.replace('"', "")
        .replace("}", "")
        .replace("{", "")
        .replace(",", "")
        .replace("\n", "")
        .replace(" ", "")
    )
